from django import forms
from django.core.paginator import EmptyPage, PageNotAnInteger, Paginator
from django.db import models
from django.utils.functional import cached_property

from wagtail.admin.panels import (
    FieldPanel,
    FieldRowPanel,
    InlinePanel,
    MultiFieldPanel,
    ObjectList,
    TabbedInterface,
)
from wagtail.blocks import ListBlock
from wagtail.contrib.settings.models import BaseSetting, register_setting
from wagtail.fields import RichTextField, StreamField
from wagtail.models import Page
from wagtail.search import index

from modelcluster.contrib.taggit import ClusterTaggableManager
from modelcluster.fields import ParentalKey
from taggit.models import ItemBase, TagBase
from wagtail_cblocks.blocks import ImageBlock

from .blocks import BodyBlock, ListLinksBlock, TextLinkBlock
from .validators import RangePageValidator


# PAGES
# ------------------------------------------------------------------------------
class HomePage(Page):
    """
    La page d'accueil du site.
    """

    info_1_title = models.CharField('titre', blank=True, max_length=128)
    info_1_text = RichTextField('texte', blank=True)
    info_2_title = models.CharField('titre', blank=True, max_length=128)
    info_2_text = RichTextField('texte', blank=True)

    quick_access = StreamField(
        [
            ('image', ImageBlock()),
        ],
        blank=True,
        verbose_name="accès rapide",
        use_json_field=True,
    )

    class Meta:
        verbose_name = "page d'accueil"
        verbose_name_plural = "pages d'accueil"

    # Attributs de Wagtail
    content_panels = Page.content_panels + [
        MultiFieldPanel(
            [
                FieldPanel('quick_access'),
            ],
            heading="Accès rapides",
        ),
        MultiFieldPanel(
            [
                FieldPanel('info_1_title'),
                FieldPanel('info_1_text'),
            ],
            heading="Information 1",
            help_text="S'affichera au dessus des brèves",
        ),
        MultiFieldPanel(
            [
                FieldPanel('info_2_title'),
                FieldPanel('info_2_text'),
            ],
            heading="Information 2",
            help_text="S'affichera au dessus des brèves",
        ),
    ]

    # N'autorise cette page qu'à la racine
    parent_page_types = ['wagtailcore.Page']

    def get_context(self, request, *args, **kwargs):
        context = super().get_context(request, *args, **kwargs)
        context['last_numeros'] = Numero.objects.live().order_by(
            '-publish_date'
        )[:5]
        context['articles'] = self.get_paginated_articles(request)
        context['breves'] = (
            Article.objects.live()
            .filter(breve=True)
            .order_by('-first_published_at')
        )
        return context

    def get_paginated_articles(self, request):
        """Retourne les derniers articles paginés"""
        page = request.GET.get('page')
        articles = (
            Article.objects.live()
            .filter(breve=False)
            .order_by('-first_published_at')
        )
        paginator = Paginator(articles, 9)
        try:
            articles = paginator.page(page)
        except PageNotAnInteger:
            articles = paginator.page(1)
        except EmptyPage:
            articles = paginator.page(paginator.num_pages)
        return articles


class AbstractBasePage(Page):
    class Meta:
        abstract = True

    show_in_menus_default = True


class StandardPage(AbstractBasePage):
    class Meta:
        verbose_name = "page standard"
        verbose_name_plural = "pages standards"

    body = StreamField(
        BodyBlock(),
        blank=True,
        verbose_name="contenu de la page",
        use_json_field=True,
    )

    content_panels = Page.content_panels + [
        FieldPanel('body'),
    ]

    search_fields = Page.search_fields + [
        index.SearchField('body'),
    ]


def get_futur_num():
    """renvoit le prochain numéro"""
    return Numero.get_last_num() + 1


class NumerosList(AbstractBasePage):
    """Liste de tous les numéros (pages enfants) du journal"""

    subpage_types = ['Numero']

    def get_context(self, request, *args, **kwargs):
        context = super().get_context(request, *args, **kwargs)
        return context

    # Attributs wagtail

    # Cette page doit être unique
    max_count = 1

    class Meta:
        verbose_name = "Liste de numéros"


class Numero(AbstractBasePage):
    """Un numero du journal"""

    # Fields

    journal_numero = models.PositiveSmallIntegerField(
        'Numéro', null=True, blank=True, unique=True, default=get_futur_num
    )
    publish_date = models.DateField(
        'Date de parution',
        null=True,
        blank=True,
        help_text="Date à laquelle le journal est paru",
    )
    image = models.ForeignKey(
        'wagtailimages.Image',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name='+',
        help_text="Image de la une du journal",
    )

    @classmethod
    def get_last_num(cls):
        """retrouve le dernier numero (retourne un integer)"""
        numeros = [
            num.journal_numero
            for num in cls.objects.all()
            if isinstance(num.journal_numero, int)
        ] + [0]
        return max(numeros)

    # wagtail attributes

    parent_page_types = ['NumerosList']
    subpage_types = ['Article']

    content_panels = Page.content_panels + [
        FieldPanel('journal_numero'),
        FieldPanel('publish_date'),
        FieldPanel('image'),
    ]


class Author(TagBase):
    """Un auteur ou un illustrateur de la brique"""

    class Meta:
        verbose_name = "Auteur"
        verbose_name_plural = "Auteurs"


class AuthorTaggedArticle(ItemBase):
    tag = models.ForeignKey(
        Author,
        verbose_name="auteurs",
        related_name="authored_articles",
        on_delete=models.CASCADE,
    )
    content_object = ParentalKey(
        to='core.Article',
        on_delete=models.CASCADE,
        related_name='authored_items',
    )


class IllustratorTaggedArticle(ItemBase):
    tag = models.ForeignKey(
        Author,
        verbose_name="auteurs",
        related_name="illustrated_articles",
        on_delete=models.CASCADE,
    )
    content_object = ParentalKey(
        to='core.Article',
        on_delete=models.CASCADE,
        related_name='illustrated_items',
    )


class Rubrique(models.Model):
    """Une catégorie pour classer les articles"""

    name = models.CharField("nom", max_length=32)

    def __str__(self):
        return self.name


class Article(StandardPage):
    """Un article du journal"""

    chapeau = RichTextField(blank=True)
    page_number = models.CharField(
        'numero(s) de page(s)',
        max_length=16,
        validators=[RangePageValidator],
        help_text=(
            "Le numéro de page au sein du journal papier. Si l'article s'étend"
            " sur plusieurs pages, vous devez les séparer avec un tiret"
        ),
    )
    image = models.ForeignKey(
        'wagtailimages.Image',
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name='+',
        help_text="Illustration de l'article",
    )
    rubrique = models.ForeignKey(
        Rubrique,
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
    )
    breve = models.BooleanField(
        'brève',
        default=False,
        help_text="Détermine si l'article est une brève ou non",
    )
    auteurs = ClusterTaggableManager(
        through='core.AuthorTaggedArticle',
        blank=True,
        verbose_name="auteurs",
        help_text="Une liste d'auteurs",
        related_name="author_articles",
    )
    illustrateurs = ClusterTaggableManager(
        through='core.IllustratorTaggedArticle',
        blank=True,
        verbose_name="illustrateurs",
        help_text="Une liste d'illustrateurs",
        related_name="illustrator_articles",
    )

    parent_page_types = ['Numero']

    content_panels = [
        FieldPanel('title'),
        FieldPanel('chapeau'),
        FieldRowPanel(
            [
                FieldPanel(
                    'page_number',
                    widget=forms.TextInput(
                        attrs={'placeholder': "exemple: 1 ou 2-5"}
                    ),
                ),
                FieldPanel('breve'),
                FieldPanel('rubrique'),
            ],
            heading="Métadonnées",
        ),
        FieldRowPanel(
            [
                FieldPanel('auteurs'),
                FieldPanel('illustrateurs'),
            ],
            heading="Auteurs",
        ),
        FieldPanel('image'),
        FieldPanel('body'),
        InlinePanel('footnotes', label="Notes de bas de page"),
    ]

    search_fields = Page.search_fields + [
        index.SearchField('body'),
    ]

    def is_void(self):
        """return True if has no editorial content"""
        return not self.chapeau and not self.body

    @cached_property
    def numero(self):
        """return the Article numero (parent page)"""
        return self.get_parent().specific

    @cached_property
    def related_articles(self):
        if self.rubrique:
            return (
                self.rubrique.article_set.live()
                .exclude(pk=self.pk)
                .order_by('-first_published_at')[:4]
            )


# SITE SETTINGS
# ------------------------------------------------------------------------------
@register_setting
class SiteSettings(BaseSetting):
    main_menu_links = StreamField(
        [('Lien', TextLinkBlock())],
        verbose_name="Liens du menu princial",
        blank=True,
        use_json_field=True,
    )
    footer_links = StreamField(
        [
            ('Liens', ListLinksBlock()),
        ],
        verbose_name="Liens du pied de page haut",
        min_num=0,
        max_num=2,
        blank=True,
        use_json_field=True,
    )
    footer_bottom_text = RichTextField(
        'Infos',
        features=['bold', 'italic'],
        blank=True,
    )
    footer_bottom_links = StreamField(
        [('Liens', ListBlock(TextLinkBlock()))],
        verbose_name="Liens du pied de page bas",
        min_num=0,
        max_num=2,
        blank=True,
        use_json_field=True,
    )

    main_menu_panels = [FieldPanel('main_menu_links')]
    footer_top_panels = [
        FieldPanel('footer_links'),
    ]
    footer_bottom_panels = [
        FieldPanel('footer_bottom_text'),
        FieldPanel('footer_bottom_links'),
    ]

    edit_handler = TabbedInterface(
        [
            ObjectList(main_menu_panels, heading='Menu principal'),
            ObjectList(footer_top_panels, heading='Pied de page haut'),
            ObjectList(footer_bottom_panels, heading='Pied de page bas'),
        ]
    )
